#include <mesh/StencilBuilder.hpp>

#include <mesh/Connectivity.hpp>
#include <mesh/ItemArray.hpp>
#include <utils/GlobalVariableManager.hpp>
#include <utils/Messenger.hpp>

template <ItemType item_type>
class StencilBuilder::Layer
{
  using ItemId = ItemIdT<item_type>;
  ItemValue<const int, item_type> m_number_of;

  std::vector<ItemId> m_item_id_vector;
  std::vector<int> m_item_number_vector;

 public:
  void
  clear()
  {
    m_item_id_vector.clear();
    m_item_number_vector.clear();
  }

  size_t
  size() const
  {
    Assert(m_item_id_vector.size() == m_item_number_vector.size());
    return m_item_id_vector.size();
  }

  const std::vector<ItemId>&
  itemIdList() const
  {
    return m_item_id_vector;
  }

  bool
  hasItemNumber(const ItemId item_id) const
  {
    const int item_number = m_number_of[item_id];

    ssize_t begin = 0;
    ssize_t end   = m_item_number_vector.size();

    while (begin < end) {
      const ssize_t mid      = (begin + end) / 2;
      const auto& mid_number = m_item_number_vector[mid];
      if (mid_number == item_number) {
        return true;   // We found the value
      } else if (mid_number < item_number) {
        if (begin == mid) {
          break;
        }
        begin = mid;
      } else {
        if (end == mid) {
          break;
        }
        end = mid;
      }
    }
    return false;
  }

  void
  add(const ItemId item_id)
  {
    const int item_number = m_number_of[item_id];

    ssize_t begin = 0;
    ssize_t end   = m_item_number_vector.size();

    while (begin < end) {
      const ssize_t mid      = (begin + end) / 2;
      const auto& mid_number = m_item_number_vector[mid];

      if (mid_number == item_number) {
        return;   // We found the value
      } else if (mid_number < item_number) {
        if (begin == mid) {
          break;
        }
        begin = mid;
      } else {
        if (end == mid) {
          break;
        }
        end = mid;
      }
    }

    m_item_id_vector.push_back(item_id);
    m_item_number_vector.push_back(item_number);

    const auto& begin_number = m_item_number_vector[begin];

    if (begin_number > item_number) {
      for (ssize_t i = m_item_number_vector.size() - 2; i >= begin; --i) {
        std::swap(m_item_number_vector[i], m_item_number_vector[i + 1]);
        std::swap(m_item_id_vector[i], m_item_id_vector[i + 1]);
      }
    } else if (begin_number < item_number) {
      for (ssize_t i = m_item_number_vector.size() - 2; i > begin; --i) {
        std::swap(m_item_number_vector[i], m_item_number_vector[i + 1]);
        std::swap(m_item_id_vector[i], m_item_id_vector[i + 1]);
      }
    }
  }

  Layer& operator=(const Layer&) = default;
  Layer& operator=(Layer&&)      = default;

  template <size_t Dimension>
  Layer(const Connectivity<Dimension>& connectivity) : m_number_of{connectivity.template number<item_type>()}
  {}

  //  Layer() = default;

  Layer(const Layer&) = default;
  Layer(Layer&&)      = default;

  ~Layer() = default;
};

template <ItemType connecting_item_type, typename ConnectivityType>
auto
StencilBuilder::_buildSymmetryConnectingItemList(const ConnectivityType& connectivity,
                                                 const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const
{
  static_assert(connecting_item_type != ItemType::cell, "cells cannot be used to define symmetry boundaries");

  using ConnectingItemId = ItemIdT<connecting_item_type>;
  ItemArray<bool, connecting_item_type> symmetry_connecting_item_list{connectivity,
                                                                      symmetry_boundary_descriptor_list.size()};
  symmetry_connecting_item_list.fill(false);

  if constexpr (ItemTypeId<ConnectivityType::Dimension>::itemTId(connecting_item_type) ==
                ItemTypeId<ConnectivityType::Dimension>::itemTId(ItemType::face)) {
    size_t i_symmetry_boundary = 0;
    for (auto p_boundary_descriptor : symmetry_boundary_descriptor_list) {
      const IBoundaryDescriptor& boundary_descriptor = *p_boundary_descriptor;

      bool found = false;
      for (size_t i_ref_connecting_item_list = 0;
           i_ref_connecting_item_list < connectivity.template numberOfRefItemList<connecting_item_type>();
           ++i_ref_connecting_item_list) {
        const auto& ref_connecting_item_list =
          connectivity.template refItemList<connecting_item_type>(i_ref_connecting_item_list);
        if (ref_connecting_item_list.refId() == boundary_descriptor) {
          found                     = true;
          auto connecting_item_list = ref_connecting_item_list.list();
          for (size_t i_connecting_item = 0; i_connecting_item < connecting_item_list.size(); ++i_connecting_item) {
            const ConnectingItemId connecting_item_id = connecting_item_list[i_connecting_item];

            symmetry_connecting_item_list[connecting_item_id][i_symmetry_boundary] = true;
          }
          break;
        }
      }
      ++i_symmetry_boundary;
      if (not found) {
        std::ostringstream error_msg;
        error_msg << "cannot find boundary '" << rang::fgB::yellow << boundary_descriptor << rang::fg::reset << '\'';
        throw NormalError(error_msg.str());
      }
    }
  } else {
    auto face_to_connecting_item_matrix =
      connectivity.template getItemToItemMatrix<ItemType::face, connecting_item_type>();
    size_t i_symmetry_boundary = 0;
    for (auto p_boundary_descriptor : symmetry_boundary_descriptor_list) {
      const IBoundaryDescriptor& boundary_descriptor = *p_boundary_descriptor;

      bool found = false;
      for (size_t i_ref_face_list = 0; i_ref_face_list < connectivity.template numberOfRefItemList<ItemType::face>();
           ++i_ref_face_list) {
        const auto& ref_face_list = connectivity.template refItemList<ItemType::face>(i_ref_face_list);
        if (ref_face_list.refId() == boundary_descriptor) {
          found = true;
          for (size_t i_face = 0; i_face < ref_face_list.list().size(); ++i_face) {
            const FaceId face_id      = ref_face_list.list()[i_face];
            auto connecting_item_list = face_to_connecting_item_matrix[face_id];
            for (size_t i_connecting_item = 0; i_connecting_item < connecting_item_list.size(); ++i_connecting_item) {
              const ConnectingItemId connecting_item_id = connecting_item_list[i_connecting_item];

              symmetry_connecting_item_list[connecting_item_id][i_symmetry_boundary] = true;
            }
          }
          break;
        }
      }
      ++i_symmetry_boundary;
      if (not found) {
        std::ostringstream error_msg;
        error_msg << "cannot find boundary '" << rang::fgB::yellow << boundary_descriptor << rang::fg::reset << '\'';
        throw NormalError(error_msg.str());
      }
    }
  }

  return symmetry_connecting_item_list;
}

template <ItemType item_type, ItemType connecting_item_type, typename ConnectivityType>
StencilArray<item_type, item_type>
StencilBuilder::_build_for_same_source_and_target(const ConnectivityType& connectivity,
                                                  const size_t& number_of_layers,
                                                  const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const
{
  if (number_of_layers == 0) {
    throw NormalError("number of layers must be greater than 0 to build stencils");
  }

  auto item_to_connecting_item_matrix = connectivity.template getItemToItemMatrix<item_type, connecting_item_type>();
  auto connecting_item_to_item_matrix = connectivity.template getItemToItemMatrix<connecting_item_type, item_type>();

  auto item_is_owned          = connectivity.template isOwned<item_type>();
  auto item_number            = connectivity.template number<item_type>();
  auto connecting_item_number = connectivity.template number<connecting_item_type>();

  using ItemId           = ItemIdT<item_type>;
  using ConnectingItemId = ItemIdT<connecting_item_type>;

  ItemArray<bool, connecting_item_type> symmetry_item_list =
    this->_buildSymmetryConnectingItemList<connecting_item_type>(connectivity, symmetry_boundary_descriptor_list);

  Array<uint32_t> row_map{connectivity.template numberOf<item_type>() + 1};
  row_map[0] = 0;
  std::vector<Array<uint32_t>> symmetry_row_map_list(symmetry_boundary_descriptor_list.size());
  for (auto&& symmetry_row_map : symmetry_row_map_list) {
    symmetry_row_map    = Array<uint32_t>{connectivity.template numberOf<item_type>() + 1};
    symmetry_row_map[0] = 0;
  }

  std::vector<ItemId> column_indices_vector;
  std::vector<std::vector<uint32_t>> symmetry_column_indices_vector(symmetry_boundary_descriptor_list.size());

  std::vector<Layer<item_type>> item_layer_list;
  std::vector<Layer<connecting_item_type>> connecting_layer_list;

  std::vector<Layer<item_type>> symmetry_item_layer_list;
  std::vector<Layer<connecting_item_type>> symmetry_connecting_layer_list;

  for (size_t i = 0; i < number_of_layers; ++i) {
    item_layer_list.emplace_back(Layer<item_type>{connectivity});
    connecting_layer_list.emplace_back(Layer<connecting_item_type>{connectivity});

    if (symmetry_boundary_descriptor_list.size() > 0) {
      symmetry_item_layer_list.emplace_back(Layer<item_type>{connectivity});
      symmetry_connecting_layer_list.emplace_back(Layer<connecting_item_type>{connectivity});
    }
  }

  for (ItemId item_id = 0; item_id < connectivity.template numberOf<item_type>(); ++item_id) {
    if (item_is_owned[item_id]) {
      for (auto&& item_layer : item_layer_list) {
        item_layer.clear();
      }
      for (auto&& connecting_layer : connecting_layer_list) {
        connecting_layer.clear();
      }

      // First layer is a special case
      {
        auto& item_layer       = item_layer_list[0];
        auto& connecting_layer = connecting_layer_list[0];

        for (size_t i_connecting_item = 0; i_connecting_item < item_to_connecting_item_matrix[item_id].size();
             ++i_connecting_item) {
          const ConnectingItemId connecting_item_id = item_to_connecting_item_matrix[item_id][i_connecting_item];

          connecting_layer.add(connecting_item_id);
        }

        for (auto connecting_item_id : connecting_layer.itemIdList()) {
          for (size_t i_item = 0; i_item < connecting_item_to_item_matrix[connecting_item_id].size(); ++i_item) {
            const ItemId layer_item_id = connecting_item_to_item_matrix[connecting_item_id][i_item];
            if (layer_item_id != item_id) {
              item_layer.add(layer_item_id);
            }
          }
        }
      }

      for (size_t i_layer = 1; i_layer < number_of_layers; ++i_layer) {
        auto& connecting_layer = connecting_layer_list[i_layer];

        auto has_connecting_item = [&i_layer, &connecting_layer_list](ConnectingItemId connecting_item_id) -> bool {
          for (ssize_t i = i_layer - 1; i >= 0; --i) {
            if (connecting_layer_list[i].hasItemNumber(connecting_item_id)) {
              return true;
            }
          }

          return false;
        };

        for (auto&& previous_layer_item_id_list : item_layer_list[i_layer - 1].itemIdList()) {
          const auto item_to_connecting_item_list = item_to_connecting_item_matrix[previous_layer_item_id_list];
          for (size_t i_connecting_item = 0; i_connecting_item < item_to_connecting_item_list.size();
               ++i_connecting_item) {
            const ConnectingItemId connecting_item_id = item_to_connecting_item_list[i_connecting_item];

            if (not has_connecting_item(connecting_item_id)) {
              connecting_layer.add(connecting_item_id);
            }
          }
        }

        auto& item_layer = item_layer_list[i_layer];

        auto has_layer_item = [&i_layer, &item_layer_list](ItemId layer_item_id) -> bool {
          for (ssize_t i = i_layer - 1; i >= 0; --i) {
            if (item_layer_list[i].hasItemNumber(layer_item_id)) {
              return true;
            }
          }

          return false;
        };

        for (auto connecting_item_id : connecting_layer.itemIdList()) {
          const auto& connecting_item_to_item_list = connecting_item_to_item_matrix[connecting_item_id];
          for (size_t i_item = 0; i_item < connecting_item_to_item_list.size(); ++i_item) {
            const ItemId layer_item_id = connecting_item_to_item_list[i_item];
            if ((layer_item_id != item_id) and (not has_layer_item(layer_item_id))) {
              item_layer.add(layer_item_id);
            }
          }
        }
      }

      for (size_t i_symmetry = 0; i_symmetry < symmetry_boundary_descriptor_list.size(); ++i_symmetry) {
        for (auto&& symmetry_item_layer : symmetry_item_layer_list) {
          symmetry_item_layer.clear();
        }
        for (auto&& symmetry_connecting_layer : symmetry_connecting_layer_list) {
          symmetry_connecting_layer.clear();
        }

        // First layer is a special case
        for (auto&& connecting_item_id : connecting_layer_list[0].itemIdList()) {
          if (symmetry_item_list[connecting_item_id][i_symmetry]) {
            symmetry_connecting_layer_list[0].add(connecting_item_id);
          }
        }

        for (auto&& connecting_item_id : symmetry_connecting_layer_list[0].itemIdList()) {
          auto item_of_connecting_item_list = connecting_item_to_item_matrix[connecting_item_id];
          for (size_t i_item_of_connecting_item = 0; i_item_of_connecting_item < item_of_connecting_item_list.size();
               ++i_item_of_connecting_item) {
            const ItemId item_id_of_connecting_item = item_of_connecting_item_list[i_item_of_connecting_item];
            symmetry_item_layer_list[0].add(item_id_of_connecting_item);
          }
        }

        for (size_t i_layer = 1; i_layer < number_of_layers; ++i_layer) {
          auto has_connecting_item = [&i_layer,
                                      &symmetry_connecting_layer_list](ConnectingItemId connecting_item_id) -> bool {
            for (ssize_t i = i_layer - 1; i >= 0; --i) {
              if (symmetry_connecting_layer_list[i].hasItemNumber(connecting_item_id)) {
                return true;
              }
            }

            return false;
          };

          for (auto&& symmetry_connecting_item_id : connecting_layer_list[i_layer].itemIdList()) {
            if (symmetry_item_list[symmetry_connecting_item_id][i_symmetry]) {
              if (not has_connecting_item(symmetry_connecting_item_id)) {
                symmetry_connecting_layer_list[i_layer].add(symmetry_connecting_item_id);
              }
            }
          }

          for (auto&& previous_layer_item_id_list : symmetry_item_layer_list[i_layer - 1].itemIdList()) {
            const auto item_to_connecting_item_list = item_to_connecting_item_matrix[previous_layer_item_id_list];
            for (size_t i_connecting_item = 0; i_connecting_item < item_to_connecting_item_list.size();
                 ++i_connecting_item) {
              const ConnectingItemId connecting_item_id = item_to_connecting_item_list[i_connecting_item];

              if (not has_connecting_item(connecting_item_id)) {
                symmetry_connecting_layer_list[i_layer].add(connecting_item_id);
              }
            }
          }

          auto has_symmetry_layer_item = [&i_layer, &symmetry_item_layer_list](ItemId layer_item_id) -> bool {
            for (ssize_t i = i_layer - 1; i >= 0; --i) {
              if (symmetry_item_layer_list[i].hasItemNumber(layer_item_id)) {
                return true;
              }
            }

            return false;
          };

          for (auto&& connecting_item_id : symmetry_connecting_layer_list[i_layer].itemIdList()) {
            auto item_of_connecting_item_list = connecting_item_to_item_matrix[connecting_item_id];
            for (size_t i_item_of_connecting_item = 0; i_item_of_connecting_item < item_of_connecting_item_list.size();
                 ++i_item_of_connecting_item) {
              const ItemId item_id_of_connecting_item = item_of_connecting_item_list[i_item_of_connecting_item];
              if (not has_symmetry_layer_item(item_id_of_connecting_item)) {
                symmetry_item_layer_list[i_layer].add(item_id_of_connecting_item);
              }
            }
          }
        }

        for (size_t i_layer = 0; i_layer < number_of_layers; ++i_layer) {
          for (auto symmetry_layer_item_id : symmetry_item_layer_list[i_layer].itemIdList()) {
            symmetry_column_indices_vector[i_symmetry].push_back(symmetry_layer_item_id);
          }
        }

        {
          size_t stencil_size = 0;
          for (size_t i_layer = 0; i_layer < number_of_layers; ++i_layer) {
            auto& item_layer = symmetry_item_layer_list[i_layer];
            stencil_size += item_layer.size();
          }

          symmetry_row_map_list[i_symmetry][item_id + 1] = symmetry_row_map_list[i_symmetry][item_id] + stencil_size;
        }
      }

      {
        size_t stencil_size = 0;
        for (size_t i_layer = 0; i_layer < number_of_layers; ++i_layer) {
          auto& item_layer = item_layer_list[i_layer];
          for (auto stencil_item_id : item_layer.itemIdList()) {
            column_indices_vector.push_back(stencil_item_id);
          }
          stencil_size += item_layer.size();
        }
        row_map[item_id + 1] = row_map[item_id] + stencil_size;
      }

    } else {
      row_map[item_id + 1] = row_map[item_id];
      for (size_t i = 0; i < symmetry_row_map_list.size(); ++i) {
        symmetry_row_map_list[i][item_id + 1] = symmetry_row_map_list[i][item_id];
      }
    }
  }

  Array<uint32_t> column_indices{column_indices_vector.size()};
  parallel_for(
    column_indices_vector.size(), PUGS_LAMBDA(size_t i) { column_indices[i] = column_indices_vector[i]; });

  ConnectivityMatrix primal_stencil{row_map, column_indices};

  typename StencilArray<item_type, item_type>::BoundaryDescriptorStencilArrayList symmetry_boundary_stencil_list;
  {
    size_t i = 0;
    for (auto&& p_boundary_descriptor : symmetry_boundary_descriptor_list) {
      symmetry_boundary_stencil_list.emplace_back(
        typename StencilArray<item_type, item_type>::
          BoundaryDescriptorStencilArray{p_boundary_descriptor,
                                         ConnectivityMatrix{symmetry_row_map_list[i],
                                                            convert_to_array(symmetry_column_indices_vector[i])}});
      ++i;
    }
  }

  return {{primal_stencil}, {symmetry_boundary_stencil_list}};
}

template <ItemType source_item_type,
          ItemType connecting_item_type,
          ItemType target_item_type,
          typename ConnectivityType>
StencilArray<source_item_type, target_item_type>
StencilBuilder::_build_for_different_source_and_target(
  const ConnectivityType& connectivity,
  const size_t& number_of_layers,
  const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const
{
  constexpr size_t Dimension = ConnectivityType::Dimension;

  if (number_of_layers == 0) {
    throw NormalError("number of layers must be greater than 0 to build stencils");
  }

  auto connecting_item_to_target_item_matrix =
    connectivity.template getItemToItemMatrix<connecting_item_type, target_item_type>();
  auto target_item_to_connecting_item_matrix =
    connectivity.template getItemToItemMatrix<target_item_type, connecting_item_type>();

  auto source_item_to_connecting_item_matrix =
    [](const auto& given_connectivity) -> ItemToItemMatrix<source_item_type, connecting_item_type> {
    if constexpr (ItemTypeId<Dimension>::itemTId(source_item_type) ==
                  ItemTypeId<Dimension>::itemTId(connecting_item_type)) {
      return {ConnectivityMatrix{}};
    } else {
      return given_connectivity.template getItemToItemMatrix<source_item_type, connecting_item_type>();
    }
  }(connectivity);

  auto source_item_is_owned   = connectivity.template isOwned<source_item_type>();
  auto target_item_number     = connectivity.template number<target_item_type>();
  auto connecting_item_number = connectivity.template number<connecting_item_type>();

  using SourceItemId     = ItemIdT<source_item_type>;
  using TargetItemId     = ItemIdT<target_item_type>;
  using ConnectingItemId = ItemIdT<connecting_item_type>;

  ItemArray<bool, connecting_item_type> symmetry_item_list =
    this->_buildSymmetryConnectingItemList<connecting_item_type>(connectivity, symmetry_boundary_descriptor_list);

  Array<uint32_t> row_map{connectivity.template numberOf<source_item_type>() + 1};
  row_map[0] = 0;
  std::vector<Array<uint32_t>> symmetry_row_map_list(symmetry_boundary_descriptor_list.size());
  for (auto&& symmetry_row_map : symmetry_row_map_list) {
    symmetry_row_map    = Array<uint32_t>{connectivity.template numberOf<source_item_type>() + 1};
    symmetry_row_map[0] = 0;
  }

  std::vector<TargetItemId> column_indices_vector;
  std::vector<std::vector<uint32_t>> symmetry_column_indices_vector(symmetry_boundary_descriptor_list.size());

  std::vector<Layer<target_item_type>> target_item_layer_list;
  std::vector<Layer<connecting_item_type>> connecting_layer_list;

  std::vector<Layer<target_item_type>> symmetry_target_item_layer_list;
  std::vector<Layer<connecting_item_type>> symmetry_connecting_layer_list;

  for (size_t i = 0; i < number_of_layers; ++i) {
    target_item_layer_list.emplace_back(Layer<target_item_type>{connectivity});
    connecting_layer_list.emplace_back(Layer<connecting_item_type>{connectivity});

    if (symmetry_boundary_descriptor_list.size() > 0) {
      symmetry_target_item_layer_list.emplace_back(Layer<target_item_type>{connectivity});
      symmetry_connecting_layer_list.emplace_back(Layer<connecting_item_type>{connectivity});
    }
  }

  for (SourceItemId source_item_id = 0; source_item_id < connectivity.template numberOf<source_item_type>();
       ++source_item_id) {
    if (source_item_is_owned[source_item_id]) {
      for (auto&& target_item_layer : target_item_layer_list) {
        target_item_layer.clear();
      }
      for (auto&& connecting_layer : connecting_layer_list) {
        connecting_layer.clear();
      }

      // First layer is a special case
      {
        auto& target_item_layer = target_item_layer_list[0];
        auto& connecting_layer  = connecting_layer_list[0];

        if constexpr (ItemTypeId<Dimension>::itemTId(source_item_type) ==
                      ItemTypeId<Dimension>::itemTId(connecting_item_type)) {
          connecting_layer.add(ConnectingItemId{static_cast<typename ConnectingItemId::base_type>(source_item_id)});
        } else {
          for (size_t i_connecting_item = 0;
               i_connecting_item < source_item_to_connecting_item_matrix[source_item_id].size(); ++i_connecting_item) {
            const ConnectingItemId connecting_item_id =
              source_item_to_connecting_item_matrix[source_item_id][i_connecting_item];

            connecting_layer.add(connecting_item_id);
          }
        }

        for (auto connecting_item_id : connecting_layer.itemIdList()) {
          for (size_t i_item = 0; i_item < connecting_item_to_target_item_matrix[connecting_item_id].size(); ++i_item) {
            const TargetItemId layer_item_id = connecting_item_to_target_item_matrix[connecting_item_id][i_item];
            target_item_layer.add(layer_item_id);
          }
        }
      }

      for (size_t i_layer = 1; i_layer < number_of_layers; ++i_layer) {
        auto& connecting_layer = connecting_layer_list[i_layer];

        auto has_connecting_item = [&i_layer, &connecting_layer_list](ConnectingItemId connecting_item_id) -> bool {
          for (ssize_t i = i_layer - 1; i >= 0; --i) {
            if (connecting_layer_list[i].hasItemNumber(connecting_item_id)) {
              return true;
            }
          }

          return false;
        };

        for (auto&& previous_layer_item_id_list : target_item_layer_list[i_layer - 1].itemIdList()) {
          const auto target_item_to_connecting_item_list =
            target_item_to_connecting_item_matrix[previous_layer_item_id_list];
          for (size_t i_connecting_item = 0; i_connecting_item < target_item_to_connecting_item_list.size();
               ++i_connecting_item) {
            const ConnectingItemId connecting_item_id = target_item_to_connecting_item_list[i_connecting_item];

            if (not has_connecting_item(connecting_item_id)) {
              connecting_layer.add(connecting_item_id);
            }
          }
        }

        auto& target_item_layer = target_item_layer_list[i_layer];

        auto has_layer_item = [&i_layer, &target_item_layer_list](TargetItemId layer_item_id) -> bool {
          for (ssize_t i = i_layer - 1; i >= 0; --i) {
            if (target_item_layer_list[i].hasItemNumber(layer_item_id)) {
              return true;
            }
          }

          return false;
        };

        for (auto connecting_item_id : connecting_layer.itemIdList()) {
          const auto& connecting_item_to_target_item_list = connecting_item_to_target_item_matrix[connecting_item_id];
          for (size_t i_item = 0; i_item < connecting_item_to_target_item_list.size(); ++i_item) {
            const TargetItemId layer_item_id = connecting_item_to_target_item_list[i_item];
            if (not has_layer_item(layer_item_id)) {
              target_item_layer.add(layer_item_id);
            }
          }
        }
      }

      for (size_t i_symmetry = 0; i_symmetry < symmetry_boundary_descriptor_list.size(); ++i_symmetry) {
        for (auto&& symmetry_target_item_layer : symmetry_target_item_layer_list) {
          symmetry_target_item_layer.clear();
        }
        for (auto&& symmetry_connecting_layer : symmetry_connecting_layer_list) {
          symmetry_connecting_layer.clear();
        }

        // First layer is a special case
        for (auto&& connecting_item_id : connecting_layer_list[0].itemIdList()) {
          if (symmetry_item_list[connecting_item_id][i_symmetry]) {
            symmetry_connecting_layer_list[0].add(connecting_item_id);
          }
        }

        for (auto&& connecting_item_id : symmetry_connecting_layer_list[0].itemIdList()) {
          auto item_of_connecting_item_list = connecting_item_to_target_item_matrix[connecting_item_id];
          for (size_t i_item_of_connecting_item = 0; i_item_of_connecting_item < item_of_connecting_item_list.size();
               ++i_item_of_connecting_item) {
            const TargetItemId item_id_of_connecting_item = item_of_connecting_item_list[i_item_of_connecting_item];
            symmetry_target_item_layer_list[0].add(item_id_of_connecting_item);
          }
        }

        for (size_t i_layer = 1; i_layer < number_of_layers; ++i_layer) {
          auto has_connecting_item = [&i_layer,
                                      &symmetry_connecting_layer_list](ConnectingItemId connecting_item_id) -> bool {
            for (ssize_t i = i_layer - 1; i >= 0; --i) {
              if (symmetry_connecting_layer_list[i].hasItemNumber(connecting_item_id)) {
                return true;
              }
            }

            return false;
          };

          for (auto&& symmetry_connecting_item_id : connecting_layer_list[i_layer].itemIdList()) {
            if (symmetry_item_list[symmetry_connecting_item_id][i_symmetry]) {
              if (not has_connecting_item(symmetry_connecting_item_id)) {
                symmetry_connecting_layer_list[i_layer].add(symmetry_connecting_item_id);
              }
            }
          }

          for (auto&& previous_layer_target_item_id_list : symmetry_target_item_layer_list[i_layer - 1].itemIdList()) {
            const auto item_to_connecting_item_list =
              target_item_to_connecting_item_matrix[previous_layer_target_item_id_list];
            for (size_t i_connecting_item = 0; i_connecting_item < item_to_connecting_item_list.size();
                 ++i_connecting_item) {
              const ConnectingItemId connecting_item_id = item_to_connecting_item_list[i_connecting_item];

              if (not has_connecting_item(connecting_item_id)) {
                symmetry_connecting_layer_list[i_layer].add(connecting_item_id);
              }
            }
          }

          auto has_symmetry_layer_item = [&i_layer,
                                          &symmetry_target_item_layer_list](TargetItemId layer_target_item_id) -> bool {
            for (ssize_t i = i_layer - 1; i >= 0; --i) {
              if (symmetry_target_item_layer_list[i].hasItemNumber(layer_target_item_id)) {
                return true;
              }
            }

            return false;
          };

          for (auto&& connecting_item_id : symmetry_connecting_layer_list[i_layer].itemIdList()) {
            auto item_of_connecting_item_list = connecting_item_to_target_item_matrix[connecting_item_id];
            for (size_t i_target_item_of_connecting_item = 0;
                 i_target_item_of_connecting_item < item_of_connecting_item_list.size();
                 ++i_target_item_of_connecting_item) {
              const TargetItemId target_item_id_of_connecting_item =
                item_of_connecting_item_list[i_target_item_of_connecting_item];
              if (not has_symmetry_layer_item(target_item_id_of_connecting_item)) {
                symmetry_target_item_layer_list[i_layer].add(target_item_id_of_connecting_item);
              }
            }
          }
        }

        for (size_t i_layer = 0; i_layer < number_of_layers; ++i_layer) {
          for (auto symmetry_layer_target_item_id : symmetry_target_item_layer_list[i_layer].itemIdList()) {
            symmetry_column_indices_vector[i_symmetry].push_back(symmetry_layer_target_item_id);
          }
        }

        {
          size_t stencil_size = 0;
          for (size_t i_layer = 0; i_layer < number_of_layers; ++i_layer) {
            auto& target_item_layer = symmetry_target_item_layer_list[i_layer];
            stencil_size += target_item_layer.size();
          }

          symmetry_row_map_list[i_symmetry][source_item_id + 1] =
            symmetry_row_map_list[i_symmetry][source_item_id] + stencil_size;
        }
      }

      {
        size_t stencil_size = 0;
        for (size_t i_layer = 0; i_layer < number_of_layers; ++i_layer) {
          auto& item_layer = target_item_layer_list[i_layer];
          for (auto stencil_item_id : item_layer.itemIdList()) {
            column_indices_vector.push_back(stencil_item_id);
          }
          stencil_size += item_layer.size();
        }
        row_map[source_item_id + 1] = row_map[source_item_id] + stencil_size;
      }

    } else {
      row_map[source_item_id + 1] = row_map[source_item_id];
      for (size_t i = 0; i < symmetry_row_map_list.size(); ++i) {
        symmetry_row_map_list[i][source_item_id + 1] = symmetry_row_map_list[i][source_item_id];
      }
    }
  }

  Array<uint32_t> column_indices{column_indices_vector.size()};
  parallel_for(
    column_indices_vector.size(), PUGS_LAMBDA(size_t i) { column_indices[i] = column_indices_vector[i]; });

  ConnectivityMatrix primal_stencil{row_map, column_indices};

  typename StencilArray<source_item_type, target_item_type>::BoundaryDescriptorStencilArrayList
    symmetry_boundary_stencil_list;
  {
    size_t i = 0;
    for (auto&& p_boundary_descriptor : symmetry_boundary_descriptor_list) {
      symmetry_boundary_stencil_list.emplace_back(
        typename StencilArray<source_item_type, target_item_type>::
          BoundaryDescriptorStencilArray{p_boundary_descriptor,
                                         ConnectivityMatrix{symmetry_row_map_list[i],
                                                            convert_to_array(symmetry_column_indices_vector[i])}});
      ++i;
    }
  }

  return {{primal_stencil}, {symmetry_boundary_stencil_list}};
}

template <ItemType source_item_type,
          ItemType connecting_item_type,
          ItemType target_item_type,
          typename ConnectivityType>
StencilArray<source_item_type, target_item_type>
StencilBuilder::_build(const ConnectivityType& connectivity,
                       const size_t& number_of_layers,
                       const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const
{
  if constexpr (connecting_item_type != target_item_type) {
    if constexpr (source_item_type == target_item_type) {
      return this
        ->_build_for_same_source_and_target<source_item_type, connecting_item_type>(connectivity, number_of_layers,
                                                                                    symmetry_boundary_descriptor_list);
    } else {
      return this->_build_for_different_source_and_target<source_item_type, connecting_item_type,
                                                          target_item_type>(connectivity, number_of_layers,
                                                                            symmetry_boundary_descriptor_list);
    }
  } else {
    std::ostringstream error_msg;
    error_msg << "cannot build stencil of " << rang::fgB::yellow << itemName(target_item_type) << rang::fg::reset
              << " using " << rang::fgB::yellow << itemName(connecting_item_type) << rang::fg::reset
              << " for connectivity";
    throw UnexpectedError(error_msg.str());
  }
}

template <ItemType source_item_type, ItemType target_item_type, typename ConnectivityType>
StencilArray<source_item_type, target_item_type>
StencilBuilder::_build(const ConnectivityType& connectivity,
                       const StencilDescriptor& stencil_descriptor,
                       const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const
{
  switch (stencil_descriptor.connectionType()) {
  case StencilDescriptor::ConnectionType::by_nodes: {
    return this->_build<source_item_type, ItemType::node, target_item_type>(connectivity,
                                                                            stencil_descriptor.numberOfLayers(),
                                                                            symmetry_boundary_descriptor_list);
  }
  case StencilDescriptor::ConnectionType::by_edges: {
    return this->_build<source_item_type, ItemType::edge, target_item_type>(connectivity,
                                                                            stencil_descriptor.numberOfLayers(),
                                                                            symmetry_boundary_descriptor_list);
  }
  case StencilDescriptor::ConnectionType::by_faces: {
    return this->_build<source_item_type, ItemType::face, target_item_type>(connectivity,
                                                                            stencil_descriptor.numberOfLayers(),
                                                                            symmetry_boundary_descriptor_list);
  }
  case StencilDescriptor::ConnectionType::by_cells: {
    return this->_build<source_item_type, ItemType::cell, target_item_type>(connectivity,
                                                                            stencil_descriptor.numberOfLayers(),
                                                                            symmetry_boundary_descriptor_list);
  }
    // LCOV_EXCL_START
  default: {
    throw UnexpectedError("invalid connection type");
  }
    // LCOV_EXCL_STOP
  }
}

CellToCellStencilArray
StencilBuilder::buildC2C(const IConnectivity& connectivity,
                         const StencilDescriptor& stencil_descriptor,
                         const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const
{
  if ((parallel::size() > 1) and
      (stencil_descriptor.numberOfLayers() > GlobalVariableManager::instance().getNumberOfGhostLayers())) {
    std::ostringstream error_msg;
    error_msg << "Stencil builder requires" << rang::fgB::yellow << stencil_descriptor.numberOfLayers()
              << rang::fg::reset << " layers while parallel number of ghost layer is "
              << GlobalVariableManager::instance().getNumberOfGhostLayers() << ".\n";
    error_msg << "Increase the number of ghost layers (using the '--number-of-ghost-layers' option).";
    throw NormalError(error_msg.str());
  }

  switch (connectivity.dimension()) {
  case 1: {
    return this->_build<ItemType::cell, ItemType::cell>(dynamic_cast<const Connectivity<1>&>(connectivity),
                                                        stencil_descriptor, symmetry_boundary_descriptor_list);
  }
  case 2: {
    return this->_build<ItemType::cell, ItemType::cell>(dynamic_cast<const Connectivity<2>&>(connectivity),
                                                        stencil_descriptor, symmetry_boundary_descriptor_list);
  }
  case 3: {
    return this->_build<ItemType::cell, ItemType::cell>(dynamic_cast<const Connectivity<3>&>(connectivity),
                                                        stencil_descriptor, symmetry_boundary_descriptor_list);
  }
  default: {
    throw UnexpectedError("invalid connectivity dimension");
  }
  }
}

NodeToCellStencilArray
StencilBuilder::buildN2C(const IConnectivity& connectivity,
                         const StencilDescriptor& stencil_descriptor,
                         const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const
{
  if ((parallel::size() > 1) and
      (stencil_descriptor.numberOfLayers() > GlobalVariableManager::instance().getNumberOfGhostLayers())) {
    std::ostringstream error_msg;
    error_msg << "Stencil builder requires" << rang::fgB::yellow << stencil_descriptor.numberOfLayers()
              << rang::fg::reset << " layers while parallel number of ghost layer is "
              << GlobalVariableManager::instance().getNumberOfGhostLayers() << ".\n";
    error_msg << "Increase the number of ghost layers (using the '--number-of-ghost-layers' option).";
    throw NormalError(error_msg.str());
  }

  switch (connectivity.dimension()) {
  case 1: {
    return this->_build<ItemType::node, ItemType::cell>(dynamic_cast<const Connectivity<1>&>(connectivity),
                                                        stencil_descriptor, symmetry_boundary_descriptor_list);
  }
  case 2: {
    return this->_build<ItemType::node, ItemType::cell>(dynamic_cast<const Connectivity<2>&>(connectivity),
                                                        stencil_descriptor, symmetry_boundary_descriptor_list);
  }
  case 3: {
    return this->_build<ItemType::node, ItemType::cell>(dynamic_cast<const Connectivity<3>&>(connectivity),
                                                        stencil_descriptor, symmetry_boundary_descriptor_list);
  }
  default: {
    throw UnexpectedError("invalid connectivity dimension");
  }
  }
}
