#ifndef STENCIL_BUILDER_HPP
#define STENCIL_BUILDER_HPP

#include <mesh/IBoundaryDescriptor.hpp>
#include <mesh/StencilArray.hpp>
#include <mesh/StencilDescriptor.hpp>

#include <string>
#include <vector>

class IConnectivity;

class StencilBuilder
{
 public:
  using BoundaryDescriptorList = std::vector<std::shared_ptr<const IBoundaryDescriptor>>;

 private:
  template <ItemType item_type>
  class Layer;

  template <ItemType connecting_item, typename ConnectivityType>
  auto _buildSymmetryConnectingItemList(const ConnectivityType& connectivity,
                                        const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const;

  template <ItemType item_type, ItemType connecting_item_type, typename ConnectivityType>
  StencilArray<item_type, item_type> _build_for_same_source_and_target(
    const ConnectivityType& connectivity,
    const size_t& number_of_layers,
    const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const;

  template <ItemType source_item_type,
            ItemType connecting_item_type,
            ItemType target_item_type,
            typename ConnectivityType>
  StencilArray<source_item_type, target_item_type> _build_for_different_source_and_target(
    const ConnectivityType& connectivity,
    const size_t& number_of_layers,
    const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const;

  template <ItemType source_item_type,
            ItemType connecting_item_type,
            ItemType target_item_type,
            typename ConnectivityType>
  StencilArray<source_item_type, target_item_type> _build(
    const ConnectivityType& connectivity,
    const size_t& number_of_layers,
    const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const;

  template <ItemType source_item_type, ItemType target_item_type, typename ConnectivityType>
  StencilArray<source_item_type, target_item_type> _build(
    const ConnectivityType& connectivity,
    const StencilDescriptor& stencil_descriptor,
    const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const;

  template <typename ConnectivityType>
  NodeToCellStencilArray _buildN2C(const ConnectivityType& connectivity,
                                   size_t number_of_layers,
                                   const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const;

  CellToCellStencilArray buildC2C(const IConnectivity& connectivity,
                                  const StencilDescriptor& stencil_descriptor,
                                  const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const;

  NodeToCellStencilArray buildN2C(const IConnectivity& connectivity,
                                  const StencilDescriptor& stencil_descriptor,
                                  const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const;

  friend class StencilManager;

  template <ItemType source_item_type, ItemType target_item_type>
  StencilArray<source_item_type, target_item_type>
  _build(const IConnectivity& connectivity,
         const StencilDescriptor& stencil_descriptor,
         const BoundaryDescriptorList& symmetry_boundary_descriptor_list) const
  {
    if constexpr ((source_item_type == ItemType::cell) and (target_item_type == ItemType::cell)) {
      return buildC2C(connectivity, stencil_descriptor, symmetry_boundary_descriptor_list);
    } else if constexpr ((source_item_type == ItemType::node) and (target_item_type == ItemType::cell)) {
      return buildN2C(connectivity, stencil_descriptor, symmetry_boundary_descriptor_list);
    } else {
      static_assert(is_false_item_type_v<source_item_type>, "invalid stencil type");
    }
  }

 public:
  StencilBuilder()                      = default;
  StencilBuilder(const StencilBuilder&) = default;
  StencilBuilder(StencilBuilder&&)      = default;
  ~StencilBuilder()                     = default;
};

#endif   // STENCIL_BUILDER_HPP
