#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <MeshDataBaseForTests.hpp>
#include <mesh/Mesh.hpp>
#include <scheme/DiscreteFunctionDPkVector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("DiscreteFunctionDPkVector", "[scheme]")
{
  SECTION("Basic interface")
  {
    const std::shared_ptr mesh_2d_v = MeshDataBaseForTests::get().cartesian2DMesh();
    const std::shared_ptr mesh_2d   = mesh_2d_v->get<Mesh<2>>();

    const std::shared_ptr mesh_2d_2_v = MeshDataBaseForTests::get().hybrid2DMesh();
    const std::shared_ptr mesh_2d_2   = mesh_2d_2_v->get<Mesh<2>>();

    DiscreteFunctionDPkVector<2, const double> R_dkp;

#ifndef NDEBUG
    REQUIRE_THROWS_WITH(R_dkp(CellId(0), 0), "DiscreteFunctionDPkVector is not built");
#endif   // NDEBUG

    {
      DiscreteFunctionDPkVector<2, double> tmp0{mesh_2d_v, 2, 3};
      DiscreteFunctionDPkVector<2, double> tmp_R_dkp(std::move(tmp0));
      tmp_R_dkp.fill(2);
      R_dkp = tmp_R_dkp;
    }

#ifndef NDEBUG
    REQUIRE_THROWS_WITH(R_dkp(CellId(0), 10), "incorrect component number");
#endif   // NDEBUG

    REQUIRE(R_dkp.degree() == 2);
    REQUIRE(R_dkp.numberOfComponents() == 3);
    REQUIRE(R_dkp.numberOfCoefficientsPerComponent() ==
            PolynomialCenteredCanonicalBasisView<2, double>::dimensionFromDegree(2));

    REQUIRE(min(R_dkp.cellArrays()) == 2);
    REQUIRE(max(R_dkp.cellArrays()) == 2);

    REQUIRE(R_dkp.dataType() == ast_node_data_type_from<double>);
    REQUIRE(R_dkp.meshVariant()->id() == mesh_2d->id());

    DiscreteFunctionDPkVector<2, double> R_dkp2;
    R_dkp2 = copy(R_dkp);

    REQUIRE(R_dkp2.degree() == 2);
    REQUIRE(R_dkp2.numberOfComponents() == 3);
    REQUIRE(R_dkp2.numberOfCoefficientsPerComponent() ==
            PolynomialCenteredCanonicalBasisView<2, double>::dimensionFromDegree(2));

    REQUIRE(min(R_dkp2.cellArrays()) == 2);
    REQUIRE(max(R_dkp2.cellArrays()) == 2);

    REQUIRE(R_dkp2.dataType() == ast_node_data_type_from<double>);
    REQUIRE(R_dkp2.meshVariant()->id() == mesh_2d_v->id());

    DiscreteFunctionDPkVector<2, double> R_dkp3(mesh_2d_2_v, 1, 2);
    R_dkp3.fill(3);

    REQUIRE(R_dkp3.degree() == 1);
    REQUIRE(R_dkp3.numberOfComponents() == 2);

    REQUIRE(min(R_dkp3.cellArrays()) == 3);
    REQUIRE(max(R_dkp3.cellArrays()) == 3);

    REQUIRE(R_dkp3.dataType() == ast_node_data_type_from<double>);
    REQUIRE(R_dkp3.meshVariant()->id() == mesh_2d_2_v->id());

    DiscreteFunctionDPkVector<2, double> R_dkp4(mesh_2d_v, R_dkp2.degree(), R_dkp2.numberOfComponents(),
                                                R_dkp2.cellArrays());

    REQUIRE(min(R_dkp4.cellArrays()) == 2);
    REQUIRE(max(R_dkp4.cellArrays()) == 2);

    R_dkp4.fill(5);
    REQUIRE(min(R_dkp4.cellArrays()) == 5);
    REQUIRE(max(R_dkp4.cellArrays()) == 5);
    REQUIRE(min(R_dkp2.cellArrays()) == 5);
    REQUIRE(max(R_dkp2.cellArrays()) == 5);

    copy_to(R_dkp, R_dkp4);
    REQUIRE(min(R_dkp4.cellArrays()) == 2);
    REQUIRE(max(R_dkp4.cellArrays()) == 2);
    REQUIRE(min(R_dkp2.cellArrays()) == 2);
    REQUIRE(max(R_dkp2.cellArrays()) == 2);

#ifndef NDEBUG
    REQUIRE_THROWS_WITH(copy_to(R_dkp, R_dkp3), "copy_to target must use the same mesh");

    DiscreteFunctionDPkVector<2, double> R_dkp5(mesh_2d, 3, 3);
    REQUIRE_THROWS_WITH(copy_to(R_dkp, R_dkp5), "copy_to target must have the same degree");
    R_dkp5 = DiscreteFunctionDPkVector<2, double>{mesh_2d, 2, 1};
    REQUIRE_THROWS_WITH(copy_to(R_dkp, R_dkp5), "copy_to target must have the same number of components");

    REQUIRE_THROWS_WITH((DiscreteFunctionDPkVector<2, double>{mesh_2d_2, 2, 3, R_dkp2.cellArrays()}),
                        "cell_array is built on different connectivity");
#endif   // NDEBUG
  }

  SECTION("1D")
  {
    constexpr size_t Dimension = 1;

    using R1 = TinyVector<1>;

    const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian1DMesh();
    const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

    auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
    auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

    const size_t degree = 4;

    DiscreteFunctionDPkVector<Dimension, double> pk(mesh_v, degree, 3);

    std::vector<double> a = {+1.0, +1.4, -6.2, +2.7, +3.1,   //
                             +2.0, -2.3, +5.4, -2.7, -1.3,   //
                             -1.2, +2.3, +3.1, +1.6, +2.3};

    for (CellId cell_id = 0; cell_id < mesh->numberOfCells(); ++cell_id) {
      auto coefficients = pk.coefficients(cell_id);
      for (size_t i = 0; i < a.size(); ++i) {
        coefficients[i] = a[i];
      }
    };

    const size_t number_of_coefs = pk.numberOfCoefficientsPerComponent();

    for (size_t i = 0; i < pk.numberOfComponents(); ++i) {
      DiscreteFunctionP0<double> p_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(const CellId cell_id) { p_xj[cell_id] = pk(cell_id, i)(xj[cell_id]); });

      REQUIRE(max(p_xj) == a[i * number_of_coefs]);
      REQUIRE(min(p_xj) == a[i * number_of_coefs]);
    }

    for (size_t i = 0; i < pk.numberOfComponents(); ++i) {
      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = 0.3 * Vj[cell_id];
          delta[cell_id]  = pk(cell_id, i)(xj[cell_id] + R1{x0})            //
                           - (a[i * number_of_coefs] +                      //
                              x0 * a[i * number_of_coefs + 1] +             //
                              x0 * x0 * a[i * number_of_coefs + 2] +        //
                              x0 * x0 * x0 * a[i * number_of_coefs + 3] +   //
                              x0 * x0 * x0 * x0 * a[i * number_of_coefs + 4]);
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
      REQUIRE(min(delta) == Catch::Approx(0.).margin(1E-14));
    }
  }

  SECTION("2D")
  {
    constexpr size_t Dimension = 2;

    using R2 = TinyVector<2>;

    const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian2DMesh();
    const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

    auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
    auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

    const size_t degree = 3;

    DiscreteFunctionDPkVector<Dimension, double> pk(mesh_v, degree, 2);

    const std::vector<double> a = {+1.0, +1.4, -6.2, +3.5, -2.3, +5.2, +6.1, +2.3, +0.5, -1.3,   //
                                   -1.3, +2.3, +2.7, +1.7, +2.1, -2.7, -5.3, +1.2, +1.3, +3.2};

    for (CellId cell_id = 0; cell_id < mesh->numberOfCells(); ++cell_id) {
      auto coefficients = pk.coefficients(cell_id);
      for (size_t i = 0; i < coefficients.size(); ++i) {
        coefficients[i] = a[i];
      }
    }

    const size_t number_of_coefs = pk.numberOfCoefficientsPerComponent();

    for (size_t i = 0; i < pk.numberOfComponents(); ++i) {
      DiscreteFunctionP0<double> p_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { p_xj[cell_id] = pk(cell_id, i)(xj[cell_id]); });

      REQUIRE(max(p_xj) == a[i * number_of_coefs]);
      REQUIRE(min(p_xj) == a[i * number_of_coefs]);
    }

    for (size_t i = 0; i < pk.numberOfComponents(); ++i) {
      DiscreteFunctionP0<double> error(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = 0.2 * Vj[cell_id];
          const double y0 = 0.3 * Vj[cell_id];
          error[cell_id]                                                //
            = std::abs(pk(cell_id, i)(xj[cell_id] + R2{x0, y0})         //
                       - (a[i * number_of_coefs]                        //
                          + x0 * a[i * number_of_coefs + 1]             //
                          + x0 * x0 * a[i * number_of_coefs + 2]        //
                          + x0 * x0 * x0 * a[i * number_of_coefs + 3]   //
                          + y0 * a[i * number_of_coefs + 4]             //
                          + y0 * x0 * a[i * number_of_coefs + 5]        //
                          + y0 * x0 * x0 * a[i * number_of_coefs + 6]   //
                          + y0 * y0 * a[i * number_of_coefs + 7]        //
                          + y0 * y0 * x0 * a[i * number_of_coefs + 8]   //
                          + y0 * y0 * y0 * a[i * number_of_coefs + 9]));
        });

      REQUIRE(max(error) == Catch::Approx(0.).margin(1E-14));
    }
  }

  SECTION("3D")
  {
    constexpr size_t Dimension = 3;

    using R3 = TinyVector<3>;

    const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian3DMesh();
    const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

    auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
    auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

    const size_t degree = 3;

    DiscreteFunctionDPkVector<Dimension, double> pk(mesh_v, degree, 1);

    const std::vector<double> a = {+1.0, +1.4, -6.2, +3.5, -2.3, +5.2, +6.1, +2.3, +0.5, -1.3,   //
                                   +2.8, -8.4, +9.5, +4.0, +4.3, +7.2, -9.1, +6.8, +6.7, +9.2};

    parallel_for(
      mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
        auto coefficients = pk.coefficients(cell_id);
        for (size_t i = 0; i < coefficients.size(); ++i) {
          coefficients[i] = a[i];
        }
      });

    DiscreteFunctionP0<double> delta_xj(mesh_v);
    parallel_for(
      mesh->numberOfCells(),
      PUGS_LAMBDA(CellId cell_id) { delta_xj[cell_id] = std::abs(pk(cell_id, 0)(xj[cell_id]) - a[0]); });

    REQUIRE(max(delta_xj) == Catch::Approx(0.).margin(1E-14));

    DiscreteFunctionP0<double> delta(mesh_v);
    parallel_for(
      mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
        const double x0 = +0.5 * Vj[cell_id];
        const double y0 = +0.3 * Vj[cell_id];
        const double z0 = -0.4 * Vj[cell_id];

        delta[cell_id] = pk(cell_id, 0)(xj[cell_id] + R3{x0, y0, z0})   //
                         - (a[0]                                        //
                            + x0 * a[1]                                 //
                            + x0 * x0 * a[2]                            //
                            + x0 * x0 * x0 * a[3]                       //
                            + y0 * a[4]                                 //
                            + y0 * x0 * a[5]                            //
                            + y0 * x0 * x0 * a[6]                       //
                            + y0 * y0 * a[7]                            //
                            + y0 * y0 * x0 * a[8]                       //
                            + y0 * y0 * y0 * a[9]                       //
                            + z0 * a[10]                                //
                            + z0 * x0 * a[11]                           //
                            + z0 * x0 * x0 * a[12]                      //
                            + z0 * y0 * a[13]                           //
                            + z0 * y0 * x0 * a[14]                      //
                            + z0 * y0 * y0 * a[15]                      //
                            + z0 * z0 * a[16]                           //
                            + z0 * z0 * x0 * a[17]                      //
                            + z0 * z0 * y0 * a[18]                      //
                            + z0 * z0 * z0 * a[19]                      //
                           );
      });

    REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
    REQUIRE(min(delta) == Catch::Approx(0.).margin(1E-14));
  }
}
